# frozen_string_literal: true

require 'forwardable'
require 'simplecov'

module LetItCrash
  module Builders
    class FileBuilder
      extend Forwardable

      def_delegators :file, :filename, :lines

      def self.build(file:, rewriter:)
        new(file: file, rewriter: rewriter).build
      end

      def initialize(file:, rewriter:)
        @file = file
        @rewriter = rewriter
      end

      def build
        Proto::File.new(
          path: path,
          digest: digest,
          lines: relevant_lines,
          coverage: coverage,
        )
      end

      private

      attr_reader :file, :rewriter

      def path
        Proto::File::Path.new(
          relative: relative,
          rewritten: rewritten,
        )
      end

      def relevant_lines
        @relevant_lines ||= lines.each_with_object({}) do |line, ret|
          number = line.line_number
          relevant = sanitizer.include?(number) && !line.skipped?
          ret[number] = line.coverage.to_i if relevant
        end
      end

      def coverage
        return 1 if relevant_lines.empty?
        relevant_lines.values.count(&:positive?) / relevant_lines.count.to_f
      end

      def relative
        pathname = Pathname.new(filename)
        pathname.absolute? ? pathname.relative_path_from(ROOT).to_s : filename
      end

      def rewritten
        rewriter ? rewriter.call(relative) : ''
      end

      def digest
        Digest::SHA256.file(filename).hexdigest
      end

      def sanitizer
        @sanitizer ||= Sanitizer.from_path(path: filename)
      end

      ROOT = Pathname.new(SimpleCov.root)
      private_constant :ROOT
    end # class FileBuilder
  end # module Builders
end # module LetItCrash
